#include <windows.h>

#include "pixelfun/pixelfun.h"
#include <math.h>

#include <time.h>
#include <stdlib.h>

#include "Vector.h"

#define rnd(n) (rand()%(n+1))
#define frnd() ((float)(rand()%2000)/2000)

#include "playmu/playmu.h"

DWORD* winbuffer;
DWORD screen[320*240];

void PutPixel(int x, int y, DWORD color)
{
	if(x<0 || y<0 || x>=320 || y>=240)
		return;
	screen[y*320+x]=color;
}

void PutPixel(PNGImage& img, int x, int y, DWORD color)
{
	if(x<0 || y<0 || x>=img.width || y>=img.height)
		return;
	img.data[y*img.width+x]=color;
}

DWORD GetPixel(PNGImage& img, int x, int y)
{
	if(x<0 || y<0 || x>=img.width || y>=img.height)
		return img.data[0];
	return img.data[y*img.width+x];
}

void DrawBar(int x0, int y0, int w, int h, DWORD color)
{
	for(int y=y0;y<y0+h;y++)
		for(int x=x0;x<x0+w;x++)
			if(x>=0 && y>=0 && x<320 && y<240)
				screen[y*320+x]=color;
}

bool clipped(int x, int y)
{
	return x<0 || y<0 || x>=320 || y>=240;
}

void DrawSprite(PNGImage& img, int x0, int y0, bool flip=false)
{
	int di=y0*320+x0;
	int si=0;
	for(int y=0;y<img.height;y++)
	{
		int fsi=si+img.width-1;
		for(int x=0;x<img.width;x++,si++,di++)
		{
			DWORD color=img.data[si];
			if(flip)
				color=img.data[fsi-x];
			if(color!=0xFF000080 && !clipped(x0+x, y0+y))
				screen[di]=color;
		}
		di+=320-img.width;
	}
}

void DrawSubSprite(PNGImage& img, int x0, int y0, int sx, int sw)
{
	int di=y0*320+x0;
	int si=sx;
	for(int y=0;y<img.height;y++)
	{
		for(int x=0;x<sw;x++,si++,di++)
		{
			DWORD color=img.data[si];
			if(color!=0xFF000080 && !clipped(x0+x, y0+y))
				screen[di]=color;
		}
		si+=img.width-sw;
		di+=320-sw;
	}
}

PNGImage img_bg;
PNGImage img_fg;
PNGImage img_cannon[10];
PNGImage img_guy[10];
PNGImage img_shot[10];
PNGImage img_bar[2];
PNGImage img_boom[10];
PNGImage img_hit[10];
PNGImage img_hp[10];
PNGImage font;

int anim_cannon_idle[]={0,-1};
int anim_cannon_hurt[]={0,0,1,1,-1};
int anim_guy_idle[]={0,-1};
int anim_guy_jump[]={1,-1};
int anim_guy_run[]={2,2,3,3,4,4,5,5,6,6,-1};
int anim_shot[]={0,0,0,1,1,1,2,2,2,-1};
int anim_shot_big[]={4,4,4,4,3,3,5,5,5,5,3,3,-1};
int anim_boom[]={2,0,1,8,1,2,3,4,4,5,5,6,6,7,-2};
int anim_hit[]={0,1,2,3,4,-2};
int anim_hp[]={0,0,0,0,0,1,1,2,2,2,1,1,-1};

void DrawNumber(int n, int x, int y, int nc=8)
{
	for(int i=0;i<nc;i++)
	{
		DrawSubSprite(font, x-i*14, y, (n%10)*14, 14);
		n/=10;
	}
}

struct Banana
{
	float x, y;
	float dx, dy;
	int health;
	PNGImage* img;
	int* anim;
	int ai;
	int at;
	int jump;
	bool onground;
	bool flip;
	int t;
	float sa;
	Banana* hook;
	int burrow;
	int type;
	int hurt;
};

const int MANY=100;
Banana cannons[MANY];
Banana shots[MANY];
Banana pofs[MANY];
Banana guy;

int curwave=1;
int score=0;
bool won=false;

void PlayAnim(Banana& joe, PNGImage* img, int* anim)
{
	if(joe.health<=0 || (joe.img==img && joe.anim==anim))
		return;
	joe.img=img;
	joe.anim=anim;
	joe.ai=0;
	joe.at=0;
}

void TickAnim(Banana& joe)
{
	if(joe.health<=0 || joe.anim==NULL)
		return;
	joe.at++;
	if(joe.at>=2)
	{
		joe.at=0;
		joe.ai++;
		if(joe.anim[joe.ai]==-1)
			joe.ai=0;
		if(joe.anim[joe.ai]==-2)
		{
			joe.ai=0;
			joe.health=0;
		}
	}
}

Banana* ShootShot(float x, float y)
{
	for(int i=0;i<MANY;i++)
	{
		if(shots[i].health<=0)
		{
			shots[i].x=x;
			shots[i].y=y;
			shots[i].health=1;
			shots[i].type=0;
			PlayAnim(shots[i], img_shot, anim_shot);
			return &shots[i];
		}
	}
	return NULL;
}

Banana* MakeBoom(float x, float y, int t)
{
	for(int i=0;i<MANY;i++)
	{
		if(pofs[i].health<=0)
		{
			pofs[i].x=x;
			pofs[i].y=y;
			pofs[i].health=1;
			pofs[i].t=t;
			pofs[i].type=0;
			pofs[i].anim=NULL;
//			PlayAnim(pofs[i], img_boom, anim_boom);
			return &pofs[i];
		}
	}
	return NULL;
}

void StartWave(int count)
{
	cannons[0].health=3;
	cannons[0].x=140;
	cannons[0].y=-40;
	cannons[0].dx=1;
	cannons[0].dy=0;
	cannons[0].t=0;
	cannons[0].sa=0.0f;
	cannons[0].burrow=0;
	cannons[0].hurt=0;
	PlayAnim(cannons[0], img_cannon, anim_cannon_idle);

	for(int i=1;i<count+1;i++)
	{
		cannons[i]=cannons[0];
		cannons[i].x=140+(float)cos((float)i*3.1415f)*(float)((i+1)/2)*40;
		cannons[i].y=-50-i*15;
		cannons[i].dx=(float)((i%2)*2-1)*(0.6f+(float)i*0.1f);
		cannons[i].t=(i*50)%120;
	}
}

void ResetGame()
{
	for(int i=0;i<MANY;i++)
	{
		cannons[i].health=0;
		shots[i].health=0;
		pofs[i].health=0;
	}
/*
	cannons[0].health=3;
	cannons[0].x=120;
	cannons[0].y=50;
	cannons[0].dx=1;
	cannons[0].dy=0;
	cannons[0].t=0;
	cannons[0].sa=0.0f;
	cannons[0].burrow=0;
	cannons[0].hurt=0;
	PlayAnim(cannons[0], img_cannon, anim_cannon_idle);

	cannons[1]=cannons[0];
	cannons[1].y=140;
	cannons[1].x=60;
	cannons[1].t=40;

	cannons[2]=cannons[0];
	cannons[2].y=170;
	cannons[2].x=230;
	cannons[2].dx=-1;
	cannons[2].t=65;

	cannons[3]=cannons[0];
	cannons[3].y=-40;
	cannons[3].x=30;
	cannons[3].t=20;
*/	

	curwave=1;
	StartWave(curwave);
	won=false;

	guy.health=10;
	guy.x=140;
	guy.y=-30;
	guy.dx=0;
	guy.dy=2;
	guy.jump=0;
	guy.onground=true;
	guy.anim=NULL;
	guy.hook=NULL;
	PlayAnim(guy, img_guy, anim_guy_idle);

	score=0;
}

void Init()
{
	playmu_init();
	playmu_loadsong("data/ld18b.smu");
	Sleep(200);
	playmu_songvolume(50);
	playmu_mastervolume(50);
	playmu_playsong(0);

	HKPtr(LoadPNG("data/bg.png", img_bg));
	HKPtr(LoadPNG("data/fg.png", img_fg));
	HKPtr(LoadPNG("data/cannon.png", img_cannon[0]));
	HKPtr(LoadPNG("data/cannon_hurt.png", img_cannon[1]));
	HKPtr(LoadPNG("data/guy.png", img_guy[0]));
	HKPtr(LoadPNG("data/guy_jump.png", img_guy[1]));
	HKPtr(LoadPNG("data/guyrun1.png", img_guy[2]));
	HKPtr(LoadPNG("data/guyrun2.png", img_guy[3]));
	HKPtr(LoadPNG("data/guyrun3.png", img_guy[4]));
	HKPtr(LoadPNG("data/guyrun4.png", img_guy[5]));
	HKPtr(LoadPNG("data/guyrun5.png", img_guy[6]));
	HKPtr(LoadPNG("data/shot1.png", img_shot[0]));
	HKPtr(LoadPNG("data/shot2.png", img_shot[1]));
	HKPtr(LoadPNG("data/shot3.png", img_shot[2]));
	HKPtr(LoadPNG("data/shot_big1.png", img_shot[3]));
	HKPtr(LoadPNG("data/shot_big2.png", img_shot[4]));
	HKPtr(LoadPNG("data/shot_big3.png", img_shot[5]));
	HKPtr(LoadPNG("data/hp1.png", img_hp[0]));
	HKPtr(LoadPNG("data/hp2.png", img_hp[1]));
	HKPtr(LoadPNG("data/hp3.png", img_hp[2]));
	HKPtr(LoadPNG("data/bar1.png", img_bar[0]));
	HKPtr(LoadPNG("data/bar2.png", img_bar[1]));
	HKPtr(LoadPNG("data/boom1.png", img_boom[0]));
	HKPtr(LoadPNG("data/boom2.png", img_boom[1]));
	HKPtr(LoadPNG("data/boom3.png", img_boom[2]));
	HKPtr(LoadPNG("data/boom4.png", img_boom[3]));
	HKPtr(LoadPNG("data/boom5.png", img_boom[4]));
	HKPtr(LoadPNG("data/boom6.png", img_boom[5]));
	HKPtr(LoadPNG("data/boom7.png", img_boom[6]));
	HKPtr(LoadPNG("data/boom8.png", img_boom[7]));
	HKPtr(LoadPNG("data/boom25.png", img_boom[8]));
	HKPtr(LoadPNG("data/hit1.png", img_hit[0]));
	HKPtr(LoadPNG("data/hit2.png", img_hit[1]));
	HKPtr(LoadPNG("data/hit3.png", img_hit[2]));
	HKPtr(LoadPNG("data/hit4.png", img_hit[3]));
	HKPtr(LoadPNG("data/hit5.png", img_hit[4]));
	HKPtr(LoadPNG("data/font.png", font));

	ResetGame();
}

void Logic()
{
	if(guy.health<=0)
	{
		if(KeyHit(Key::Enter) || KeyHit(Key::Space) || KeyHit(Key::Z))
			ResetGame();
		return;
	}

	bool run=false;
	if(KeyPressed(Key::Right))
	{
		guy.dx+=0.4f;
		guy.flip=false;
		if(guy.hook!=NULL)
			guy.hook->dx+=0.04f;
		run=true;
	}
	if(KeyPressed(Key::Left))
	{
		guy.dx-=0.4f;
		guy.flip=true;
		if(guy.hook!=NULL)
			guy.hook->dx-=0.04f;
		run=true;
	}
	guy.dx*=0.8f;
	if(fabs(guy.dx)>2.0f)
		guy.dx*=2.0f/fabs(guy.dx);
	if(fabs(guy.dx)<0.1f)
		guy.dx=0.0f;
	guy.x+=guy.dx;
	if(guy.x<16)
	{
		guy.x=16;
		guy.dx=0;
	}
	if(guy.x>283)
	{
		guy.x=283;
		guy.dx=0;
	}
	if(KeyHit(Key::Up))
	{
		if(guy.onground)
		{
			guy.jump=25;
			guy.dy-=2.0f;
			if(guy.dy<-3.0f)
				guy.dy=-3.0f;
			guy.hook=NULL;
		}
	}
	if(!KeyPressed(Key::Up))
		guy.jump=0;
	static float guypy=0;
	static float guypx=0;
	guy.onground=false;
	guy.dy+=0.15f;
	if(guy.dy>5.0f)
		guy.dy=5.0f;
	if(guy.jump>0)
	{
		guy.jump--;
		guy.dy-=0.18f;
	}
	guy.y+=guy.dy;
	if(guy.y>175)
	{
		guy.y=175;
		guy.jump=0;
		guy.dy=0;
		guy.onground=true;

		if(run)
			PlayAnim(guy, img_guy, anim_guy_run);
		else
			PlayAnim(guy, img_guy, anim_guy_idle);
	}
	if(!guy.onground)
		PlayAnim(guy, img_guy, anim_guy_jump);

/*	static tt=0;
	tt++;
	tt%=40;
	if(tt==0)
	{
		Banana* shot=ShootShot(guy.x+6, guy.y-5);
		PlayAnim(*shot, img_shot, anim_shot_big);
		shot->dx=0;
		shot->dy=-3;
		shot->type=1;
	}
*/
	static int enfd=0;
	enfd++;
	enfd%=3;

	int enalive=0;
	for(int i=0;i<MANY;i++)
	{
		if(cannons[i].health<=0)
			continue;

		enalive++;

		if(won && cannons[i].y<-40)
		{
			score+=15000;
			curwave++;
			StartWave(curwave);
			won=false;
		}

		if(cannons[i].burrow)
		{
			cannons[i].t=1;
			cannons[i].burrow--;
			if(cannons[i].burrow==0)
			{
				cannons[i].dy=-2;
				MakeBoom(cannons[i].x-10, cannons[i].y-20, 0);
				MakeBoom(cannons[i].x-10, cannons[i].y-10, -5);
				MakeBoom(cannons[i].x-10, cannons[i].y, -10);
			}
			if(cannons[i].burrow==120 || cannons[i].burrow==90 || cannons[i].burrow==60)
			{
				Banana* shot=ShootShot(cannons[i].x+11, cannons[i].y);
				PlayAnim(*shot, img_shot, anim_shot_big);
				shot->dx=0;
				shot->dy=-3;
				shot->type=1;
				MakeBoom(cannons[i].x-10, cannons[i].y-25, 0);
			}
		}
		else
		{
			if(!won && guy.hook==NULL && cannons[i].burrow==0 && !guy.onground && guy.dy>0 && guy.x>cannons[i].x-13 && guy.x<cannons[i].x+18 && guy.y>cannons[i].y-25 && guy.y<cannons[i].y-17)
			{
				guy.t=0;
				guy.hook=&cannons[i];
				cannons[i].dy+=guy.dy*0.3f;
				cannons[i].dy+=(guy.dy*0.6f-cannons[i].dy)*0.5f;
			}
			
			if(cannons[i].hurt>0)
			{
				cannons[i].hurt--;
				PlayAnim(cannons[i], img_cannon, anim_cannon_hurt);
			}
			else
				PlayAnim(cannons[i], img_cannon, anim_cannon_idle);
	
			if(cannons[i].y<80 || cannons[i].dy<-1.5f)
				cannons[i].dy+=0.02f;
			if(cannons[i].y>120 || cannons[i].dy>1.5f)
				cannons[i].dy-=0.02f;
			if(cannons[i].x<16)
				cannons[i].dx=fabs(cannons[i].dx);
			if(cannons[i].x>275)
				cannons[i].dx=-fabs(cannons[i].dx);
			if(cannons[i].dx>1.0f)
				cannons[i].dx+=(1.0f-cannons[i].dx)*0.05f;
			if(cannons[i].dx<-1.0f)
				cannons[i].dx+=(-1.0f-cannons[i].dx)*0.05f;
			if(won)
				cannons[i].dy-=0.03f;
			cannons[i].x+=cannons[i].dx;
			cannons[i].y+=cannons[i].dy;
			if(cannons[i].y>195)
			{
				cannons[i].y=195;
				if(cannons[i].dy>1.5f)
				{
					cannons[i].burrow=180;
					if(guy.hook=&cannons[i])
						guy.hook=NULL;
					MakeBoom(cannons[i].x-10, cannons[i].y-19, 0);
					MakeBoom(cannons[i].x-20, cannons[i].y-15, -5);
					MakeBoom(cannons[i].x, cannons[i].y-15, -7);
					score+=2500;
				}
				cannons[i].dy=0;
			}

			if(enfd<2)
				cannons[i].t++;
			cannons[i].t%=120;
			Vector cp=Vector(cannons[i].x+15, cannons[i].y+15, 0);
			Vector gp=Vector(guy.x+10, guy.y+15, 0);
			if(cannons[i].t==0 && cannons[i].y>10 && guy.hook!=&cannons[i] && (cp-gp).Length()>60)
			{
				cannons[i].t++;
				Banana* shot=ShootShot(cannons[i].x+12, cannons[i].y+24);
				Vector pp=Vector(guy.x+10, guy.y+10, 0);
				Vector sp=Vector(shot->x+4, shot->y+4, 0);
				Vector d=pp-sp;
				d.NormalizeSafe();
				d*=1.5f;
				shot->dx=d.x;
				shot->dy=d.y;

				MakeBoom(cannons[i].x-10, cannons[i].y-10, 0);
			}
		}
	}

	if(enalive<=1)
		won=true;

	for(int i=0;i<MANY;i++)
	{
		if(shots[i].health<=0)
			continue;
		shots[i].x+=shots[i].dx;
		shots[i].y+=shots[i].dy;
		if(shots[i].x<-20 || shots[i].x>340 || shots[i].y<-40 || shots[i].y>280)
			shots[i].health=0;
		Vector sp=Vector(shots[i].x+4, shots[i].y+4, 0);
		Vector gp=Vector(guy.x+10, guy.y+15, 0);
		if((sp-gp).Length()<10)
		{
			shots[i].health=0;
			Vector r=gp-sp;
			r.NormalizeSafe();
			r.y*=0.5f;
			r*=2.5f;
			guy.dx=r.x;
			guy.dy=r.y;
			guy.onground=false;
			guy.hook=NULL;
			guy.health-=shots[i].type==1?3:1;
			Banana* boom=MakeBoom(shots[i].x-2, shots[i].y-2, 1);
			PlayAnim(*boom, img_hit, anim_hit);
		}
		for(int j=0;j<MANY;j++)
		{
			if(cannons[j].health<=0 || cannons[j].burrow>0 || shots[i].type==0)
				continue;
			Vector cp=Vector(cannons[j].x+15, cannons[j].y+15, 0);
			if((sp-cp).Length()<20)
			{
				cannons[j].health--;
				if(cannons[j].health<=0)
				{
					score+=6500;
					Banana* hp=MakeBoom(cannons[j].x+10, cannons[j].y+10, 0);
					hp->type=1;
					hp->t=1;
					hp->dx=0;
					hp->dy=-1.5f;
					PlayAnim(*hp, img_hp, anim_hp);
				}
//				{
					MakeBoom(cannons[j].x-10, cannons[j].y-10, 0);
					MakeBoom(cannons[j].x-10+5, cannons[j].y+12, -6);
					MakeBoom(cannons[j].x-10-4, cannons[j].y-2, -9);
					MakeBoom(cannons[j].x-10, cannons[j].y-10, -15);
					MakeBoom(cannons[j].x-10+5, cannons[j].y-12, -20);
//				}
				cannons[j].hurt=30;
				cannons[j].dy=-1.5f;
				cannons[j].y-=7;
				shots[i].health=0;
				score+=1250;
				break;
			}
		}
	}

	if(guy.hook!=NULL)
	{
		if(guy.hook->health<=0)
			guy.hook=NULL;
		else
		{
			PlayAnim(guy, img_guy, anim_guy_jump);
			guy.t++;
			float f=(float)guy.t/10;
			if(f>1.0f) f=1.0f;
			guy.x+=((guy.hook->x+5)-guy.x)*f;
			guy.y+=((guy.hook->y-25)-guy.y)*f;
			guy.onground=true;
			guy.dy=guy.y-guypy;
			guy.dx=guy.x-guypx;
			guy.hook->t=1;
		}
	}
	guypy=guy.y;
	guypx=guy.x;

	for(int i=0;i<MANY;i++)
	{
		if(pofs[i].health<=0)
			continue;
		if(pofs[i].t==0)
			PlayAnim(pofs[i], img_boom, anim_boom);
		pofs[i].t++;
		
		if(pofs[i].type==1)
		{
			if(pofs[i].t>240)
				pofs[i].health=0;
			pofs[i].dy+=0.06f;
			pofs[i].y+=pofs[i].dy;
			if(pofs[i].y>195)
			{
				pofs[i].y=195;
				pofs[i].dy=0;
			}
			Vector hp=Vector(pofs[i].x+4, pofs[i].y+4, 0);
			Vector gp=Vector(guy.x+10, guy.y+15, 0);
			if((hp-gp).Length()<15 && pofs[i].y>190)
			{
				pofs[i].health=0;
				guy.health+=2;
				if(guy.health>10)
					guy.health=10;
			}
		}
	}

	TickAnim(guy);
	for(int i=0;i<MANY;i++)
	{
		TickAnim(cannons[i]);
		TickAnim(shots[i]);
		TickAnim(pofs[i]);
	}
}

void Render()
{
	DrawSprite(img_bg,0,0);

	for(int i=0;i<MANY;i++)
		if(shots[i].health>0)
			DrawSprite(shots[i].img[shots[i].anim[shots[i].ai]], shots[i].x, shots[i].y);

	DrawSprite(guy.img[guy.anim[guy.ai]], guy.x, guy.y, guy.flip);

	for(int i=0;i<MANY;i++)
		if(cannons[i].health>0)
			DrawSprite(cannons[i].img[cannons[i].anim[cannons[i].ai]], cannons[i].x, cannons[i].y);

	DrawSprite(img_fg,0,0);

	for(int i=0;i<MANY;i++)
		if(pofs[i].health>0 && pofs[i].anim!=NULL)
			DrawSprite(pofs[i].img[pofs[i].anim[pofs[i].ai]], pofs[i].x, pofs[i].y);

	for(int i=0;i<10;i++)
		DrawSprite(img_bar[i<guy.health?0:1], 20+i*6, 5);

	DrawNumber(score, 290, 5);
	DrawNumber(curwave, 152, 213, 2);

	static DWORD line[640];
	int di=0;
	int si=0;
	for(int y=0;y<480;y+=2)
	{
		for(int x=0;x<640;x+=2)
		{
			line[x]=screen[si];
			line[x+1]=screen[si++];
		}
		if(guy.health<=0)
		{
			for(int x=0;x<640;x++)
			{
				DWORD& c=line[x];
				c=((c>>16)&0xFF)+((c>>8)&0xFF)+(c&0xFF);
				c/=4;
				c+=40;
				DWORD b=c/2;
				DWORD d=c/4;
				c=(c<<16)|(d<<8)|b;
			}
		}
		for(int x=0;x<640;x++)
			winbuffer[di++]=line[x];
		for(int x=0;x<640;x++)
			winbuffer[di++]=line[x];
	}
}

void main()
{
	srand(time(NULL));
	
	winbuffer=OpenWindow("ld18", 640, 480);
	if(winbuffer==NULL)
		return;

	Init();

	for(;;)
	{
		if(KeyPressed(Key::Q) || WindowClosed())
			break;

		Logic();
		Render();

		ForgetKeys();

		UpdateWindow();
		
		Sleep(7);
	}

	playmu_close();
	
	HKFree();
	CloseWindow();
}

